/**
* Copyright © [2024] [GST378]. All rights reserved.
* Personal use only. Modifications and redistribution are prohibited.
* See LICENSE file for details.
*/
import { world, system, ItemStack, GameMode } from '@minecraft/server';
import { ADDON_CONFIG } from './../config/addonConfig.js';
import PaintingManager from './PaintingManager.js';
import PaintModeManager from './PaintModeManager.js';
import PlayerManager from './Player/PlayerManager.js';
import { EaselForm, DeletePaintingForm } from './../forms/EaselForm.js';
import { OBJECTIVES, setActionScore } from './../scores.js';
import { ENTITIES, ITEMS, TAGS, DYNAMIC_PROPERTY_IDS, PARTICLES } from './../data.js';
import { spawnPlayerControl } from './PaintMode/index.js';

//function convertBytes(bytes) { return bytes >= 1e6 ? (bytes / 1e6).toFixed(2) + " MB" : (bytes / 1e3).toFixed(2) + " KB"; }

class EaselManager {
    
    static openForm(player, easel, canvas) {
        EaselForm(canvas).show(player).then((e) => {
            if (e.canceled) return;
            const isSaved = canvas?.hasTag(TAGS.IS_SAVED) ?? false;
            if (e.selection === 0) {
                isSaved ? EaselManager.dropPainting(easel, canvas) 
                : PaintingManager.savePainting(player, canvas);
            } else if (e.selection === 1) {
                EaselManager[isSaved ? 'dropPainting' : 'deleteCanvas'](easel, canvas, player, false);
            }
        });
    }
    
    static openFrameForm(player, easel, canvas) {
        EaselForm(canvas).show(player).then((e) => {
            if (e.canceled) return;
            if (e.selection === 0) EaselManager.dropPainting(easel, canvas);
        });
    }
    
    static addCanvas(easel, player) {
        if (!easel || !player) return;
        const rotations = { 'north': 180, 'east': 270, 'south': 0, 'west': 90 };
        const { x, y, z } = easel.bottomCenter();
        const cardinalDirection = easel.permutation.getState('minecraft:cardinal_direction');
        easel.dimension.runCommand(`summon ${ENTITIES.CANVAS} ${x} ${y} ${z} ${rotations[cardinalDirection]} 0`);
        player.playSound('block.itemframe.place');
        PlayerManager.decrementStack(player);
    }
    
    static addPainting(easel, player, blockType = 'easel') {
        if (!easel || !player) return;
        const paintingItem = PlayerManager.getItem(player, 'mainhand');
        //console.warn(convertBytes(paintingItem.getDynamicPropertyTotalByteCount()));
        const paintingTitle = paintingItem.nameTag.replace('§r§f', '');
        const author = paintingItem.getLore()[0].replace('§r§7', '');
        const rotations = { 'north': 180, 'east': 270, 'south': 0, 'west': 90 };
        const { x, y, z } = easel.bottomCenter();
        const cardinalDirection = blockType === 'easel' ? easel.permutation.getState('minecraft:cardinal_direction') : easel.permutation.getState('minecraft:block_face');
        easel.dimension.runCommand(`summon ${blockType === 'easel' ? ENTITIES.CANVAS : ENTITIES.FRAME_CANVAS} ${x} ${y} ${z} ${rotations[cardinalDirection]} 0`);
        const canvas = EaselManager.getCanvas(easel);
        if (!canvas) return;
        system.runTimeout(() => { PaintingManager.loadPaintingColors(paintingItem, canvas); }, 2);
        canvas.addTag(TAGS.IS_SAVED);
        canvas.addTag(TAGS.PAINTING_TITLE + '|' + paintingTitle);
        canvas.addTag(TAGS.AUTHOR + '|' + author);
        player.playSound('block.itemframe.place');
        PlayerManager.decrementStack(player, 1);
    }
    
    static dropPainting(easel, canvas) {
        if (!easel || !canvas) return;
        const tags = canvas.getTags();
        const paintingTitle = tags.find(tag => tag.startsWith(TAGS.PAINTING_TITLE))?.split('|')[1] ?? '---';
        const author = tags.find(tag => tag.startsWith(TAGS.AUTHOR))?.split('|')[1] ?? '---';
        const paintingItem = new ItemStack(ITEMS.PAINTING);
        paintingItem.nameTag = '§r§f' + paintingTitle;
        paintingItem.setLore([`§r§7${author}`]);
        const colors = PaintingManager.getPaintingColors(canvas);
        paintingItem.setDynamicProperty(DYNAMIC_PROPERTY_IDS.COLORS, colors);
        canvas.remove();
        easel.dimension.playSound('block.itemframe.break', easel.center());
        easel.dimension.spawnItem(paintingItem, easel.center());
    }
    
    static deleteCanvas(easel, canvas, player, shouldDelete = false) {
        if (!easel || !canvas) return;
        if (shouldDelete || !player?.isValid()) {
            canvas.remove();
            easel.dimension.playSound('block.itemframe.break', easel.center());
            easel.dimension.spawnParticle(PARTICLES.DELETE, easel.center());
            easel.dimension.spawnItem(new ItemStack(ITEMS.CANVAS, 1), easel.center());
        } else {
            DeletePaintingForm().show(player).then((formData) => {
                if (formData.canceled || formData.selection === 1) return;
                else EaselManager.deleteCanvas(easel, canvas, player, true);
            });
        }
    }
    
    static getCanvas(easel) {
        if (!easel) return;
        return easel.dimension.getEntities({ families: [`${ADDON_CONFIG.NAMESPACE}.canvas`], location: easel.bottomCenter(), maxDistance: 0.5, closest: 1 })[0];
    }
    
    static startPaintMode(player, easel, canvas, mainhand) {
        if (!easel || !player || !canvas) return;
        if (canvas.hasTag(TAGS.IS_SAVED) && mainhand?.typeId !== ITEMS.EYEDROPPER) return;
        if (OBJECTIVES.ACTIONS?.getScore(player) ?? 0 > 0) return;
        setActionScore(player, 'start_paint_mode');
        const playerControl = spawnPlayerControl(easel, player);
        easel.setPermutation(easel.permutation.withState(`${ADDON_CONFIG.NAMESPACE}:paint_mode`, true));
        new PaintModeManager(player, playerControl, easel, canvas).managePaintMode(canvas.hasTag(TAGS.IS_SAVED));
    }
}

export default EaselManager;