/**
* Copyright © [2024] [GST378]. All rights reserved.
* Personal use only. Modifications and redistribution are prohibited.
* See LICENSE file for details.
*/
import { system } from '@minecraft/server';
import { uiManager } from '@minecraft/server-ui';
import { ADDON_CONFIG } from './../../../config/addonConfig.js';
import { PALETTES, CUSTOM_PALETTES, hexToRGB, getHexValue, rgbToHsb, hsbToRGB } from './../../../config/colorsConfig.js';
import { getViewPointForPalette } from './viewPoints.js';
import { PaletteForm, ManageCustomPalettes, AddCustomPalettes, DeleteCustomPalettes, PickColorForm } from './../../../forms/PaletteForm.js';
import { OBJECTIVES } from './../../../scores.js';
import { ENTITIES, PROPERTY_IDS, packRGB, unPackRGB } from './../../../data.js';

const _paletteOffset = {
    'north': { x: 0.5, y: 0.5, z: -0.5 },
    'east': { x: 1.5, y: 0.5, z: 0.5 },
    'south': { x: 0.5, y: 0.5, z: 1.5 },
    'west': { x: -0.5, y: 0.5, z: 0.5 }
};

function _updatePalette(palette, rgb) {
    palette.setProperty(PROPERTY_IDS.NEW_COLOR, rgb);
    const HSB = rgbToHsb(unPackRGB(rgb));
    palette.setProperty(PROPERTY_IDS.POS_X, HSB.s);
    palette.setProperty(PROPERTY_IDS.POS_Y, HSB.b);
    const hueRGB = hsbToRGB(HSB.h, 255, 255);
    palette.setProperty(PROPERTY_IDS.HUE_COLOR, packRGB(hueRGB.r, hueRGB.g, hueRGB.b));
    palette.setProperty(PROPERTY_IDS.HUE, HSB.h);
}

function spawnPalette(easel, cardinalDirection, currentColor) {
    const offset = _paletteOffset[cardinalDirection];
    const palette = easel.dimension.spawnEntity(ENTITIES.PALETTE_UI, { x: easel.location.x + offset.x, y: easel.location.y + offset.y, z: easel.location.z + offset.z });
    palette.triggerEvent(ADDON_CONFIG.NAMESPACE + ':spawn_' + cardinalDirection);
    palette.setProperty(PROPERTY_IDS.CURRENT_COLOR, currentColor);
    _updatePalette(palette, currentColor);
    return palette;
}

function selectPaletteColor(player, palette, point, paletteType = 0, cardinalDirection) {
    if (paletteType === 0) {
        const { x , y } = getViewPointForPalette(player, cardinalDirection, 35, 45, -5, { x: 256, y: 295 });
        if (point.y > 5) { palette.setProperty(PROPERTY_IDS.POS_X, x); palette.setProperty(PROPERTY_IDS.POS_Y, Math.max(0, Math.min(Math.floor(y + y*0.2), 255))); }
        else if (point.y >= 4) { palette.setProperty(PROPERTY_IDS.HUE, x); }
        else if (point.y >= 2 && x < 50) { pickHexColor(player, palette); }
        else if (point.y >= 2 && x < 115) { _updatePalette(palette, palette.getProperty(PROPERTY_IDS.CURRENT_COLOR)); }
        else if (point.y >= 2 && x > 210) { manageCustomPalettes(player, palette); }
        system.run(() => {
            if (!palette?.isValid()) return;
            const hueRGB = hsbToRGB(palette.getProperty(PROPERTY_IDS.HUE), 255, 255);
            palette.setProperty(PROPERTY_IDS.HUE_COLOR, packRGB(hueRGB.r, hueRGB.g, hueRGB.b));
            const newColor = hsbToRGB(palette.getProperty(PROPERTY_IDS.HUE), palette.getProperty(PROPERTY_IDS.POS_X), palette.getProperty(PROPERTY_IDS.POS_Y));
            palette.setProperty(PROPERTY_IDS.NEW_COLOR, packRGB(newColor.r, newColor.g, newColor.b));
        });
    } else if (paletteType === 1) {
        system.runTimeout(() => { if (palette?.isValid()) { palette.setProperty(PROPERTY_IDS.PALETTE_TYPE, 0); } }, 3);
        point.x = (point.x <= 45) ? 0 : (point.x <= 127) ? 1 : (point.x <= 210) ? 2 : 3;
        const selectedColor = point.x + Math.floor(point.y/4) * 4;
        _updatePalette(palette, palette.getProperty(ADDON_CONFIG.NAMESPACE + ':color_' + selectedColor));
    }
}

function choosePaletteColor(player, palette) {
    const RGB = unPackRGB(palette.getProperty(PROPERTY_IDS.NEW_COLOR));
    OBJECTIVES.COLOR_RGB.setScore(player, packRGB(RGB.r, RGB.g, RGB.b));
    palette.remove();
}

function pickHexColor(player, palette) {
    PickColorForm().show(player).then((formData) => {
        if (formData.canceled || !palette?.isValid()) return;
        uiManager.closeAllForms(player);
        const RGB = hexToRGB(getHexValue(formData.formValues[0]));
        _updatePalette(palette, packRGB(RGB.r, RGB.g, RGB.b));
    });
}

function loadPaletteColors(palette, paletteIndex = 0) {
    let colorIndex = 0;
    palette.setProperty(PROPERTY_IDS.PALETTE_TYPE, 1);
    if (paletteIndex < Object.keys(PALETTES).length) {
        for (const hex of PALETTES[paletteIndex]) {
            const RGB = hexToRGB(hex);
            palette.setProperty(ADDON_CONFIG.NAMESPACE + ':color_' + colorIndex, packRGB(RGB.r, RGB.g, RGB.b));
            colorIndex++
        }
    } else {
        paletteIndex = paletteIndex - Object.keys(PALETTES).length;
        for (const hex of Object.values(CUSTOM_PALETTES)[paletteIndex]) {
            const RGB = hexToRGB(hex);
            palette.setProperty(ADDON_CONFIG.NAMESPACE + ':color_' + colorIndex, packRGB(RGB.r, RGB.g, RGB.b));
            colorIndex++
        }
    }
}

function manageCustomPalettes(player, palette) {
    PaletteForm().show(player).then((formData) => {
        if (formData.canceled || !palette?.isValid()) return;
        uiManager.closeAllForms(player);
        if (formData.selection < (Object.keys(PALETTES).length + Object.keys(CUSTOM_PALETTES).length)) {
            loadPaletteColors(palette, formData.selection);
            palette.setProperty(PROPERTY_IDS.PALETTE_TYPE, 1);
        } else {
            ManageCustomPalettes().show(player).then((formData) => {
                if (formData.canceled) return;
                uiManager.closeAllForms(player);
                const customPalettesCount = Object.keys(CUSTOM_PALETTES).length;
                if (formData.selection === customPalettesCount) {
                    if (customPalettesCount < 10) AddCustomPalettes(player);
                }
                else { DeleteCustomPalettes(player, formData.selection); }
            });
        }
    });
}

export { spawnPalette, selectPaletteColor, choosePaletteColor }