/**
* Copyright © [2024] [GST378]. All rights reserved.
* Personal use only. Modifications and redistribution are prohibited.
* See LICENSE file for details.
*/
import { system, GameMode, HudElement, HudVisibility } from '@minecraft/server';
import { ADDON_CONFIG } from './../config/addonConfig.js';
import { PAINT_MODE_CONFIG } from './../config/paintModeConfig.js';
import { SHAPES, allShapes } from './../config/shapesConfig.js';
import { ShapesForm } from './../forms/ShapesForm.js';
import { ExposureForm } from './../forms/ExposureForm.js';
import PaintingManager from './PaintingManager.js';
import PlayerManager from './Player/PlayerManager.js';
import { OBJECTIVES, getPaintingAction, setActionScore } from './../scores.js';
import { ITEMS, TAGS, EVENTS, PIXELS, PROPERTY_IDS, DYNAMIC_PROPERTY_IDS } from './../data.js';
import {
    spawnPlayerControl,
    getViewPoint, getViewPointHorizontal,
    getPixel, getPixelColor, clearPixel, paintPixel, fillPixels, changePixelBrightness,
    spawnPalette, selectPaletteColor, choosePaletteColor,
    createShape, spawnShapeBox
} from './PaintMode/index.js';

const VIEW_POINTS = {
    FOV: { x: 60, y: 60 },
    OFFSET: { x: 0, y: -5 }
};

class PaintModeManager {
    constructor(player, playerControl, easel, canvas) {
        this.point = { x: 0, y: 0 };
        this.horizontalPoint = 0;
        this.paintingAction = 'start_paint_mode';
        this.player = player;
        this.playerControl = playerControl;
        this.easel = easel;
        this.canvas = canvas;
        this.pixels = easel.dimension.getEntities({ location: easel.bottomCenter(), families: [`${ADDON_CONFIG.NAMESPACE}.pixels`], closest: 16 });
        this.palette = null;
        this.brushSize = 1;
        this.colorRGB = 0;
        this.paintingSteps = [];
        this.currentStep = null;
        this.currentIndex = -1;
        this.cardinalDirection = easel.permutation.getState('minecraft:cardinal_direction');
        this.selectedShape = allShapes[0];
        this.shapePoint = { from: { x: 0, y: 0 }, to: { x: 0, y: 0 } };
        this.shapeBox = null;
        this.shapeStep = 0;
        this.exposureStep = 0;
        this.exposureValue = 30;
        this.lightenOrDarken = false;
        this.exposurePoints = [];
    }
    
    static resetPlayer(player) {
        if (!player || !player.isValid()) return;
        OBJECTIVES.ACTIONS.setScore(player, 0);
        player.onScreenDisplay.setHudVisibility(HudVisibility.Reset, [HudElement.HorseHealth, HudElement.ItemText, HudElement.Crosshair]);
    }
    // resetPlayerControl() { this.playerControl = spawnPlayerControl(this.easel, this.player); }
    getBrushSize(mainhand) { return mainhand?.hasTag(TAGS.BRUSH_SIZE + '3') ? 3 : 1; }
    
    updateSteps(actionScore = 'start_paint_mode') {
        this.resetActions(actionScore);
        if (PAINT_MODE_CONFIG.STEP_COUNT.defaultValue === 0) return;
        if (this.currentIndex < this.paintingSteps.length - 1) this.paintingSteps.splice(this.currentIndex + 1, this.paintingSteps.length - this.currentIndex - 1);
        if (this.paintingSteps.length === PAINT_MODE_CONFIG.STEP_COUNT.defaultValue) this.paintingSteps.shift();
        this.currentStep = PaintingManager.getPaintingColors(this.canvas);
        this.paintingSteps.push(this.currentStep);
        this.currentIndex = this.paintingSteps.length - 1;
    }
    
    undoRedo(mode = 'undo') {
        setActionScore(this.player, 'start_paint_mode');
        if (PAINT_MODE_CONFIG.STEP_COUNT.defaultValue === 0) return;
        if (mode === 'undo' && this.currentIndex > 0) this.currentIndex--;
        else if (mode === 'redo' && this.currentIndex < this.paintingSteps.length - 1) this.currentIndex++;
        else return;
        this.currentStep = this.paintingSteps[this.currentIndex];
        PaintingManager.loadStep(this.canvas, this.currentStep);
    }
    
    resetActions(actionScore = 'start_paint_mode') {
        this.shapeStep = 0;
        this.exposurePoints = [];
        this.paintingAction = actionScore;
        if (this.palette?.isValid()) { choosePaletteColor(this.player, this.palette); return; }
        if (this.shapeBox?.isValid()) { this.shapeBox.remove(); }
        setActionScore(this.player, actionScore);
    }
    
    managePalette() {
        if (!this.palette?.isValid()) { this.palette = spawnPalette(this.easel, this.cardinalDirection, this.colorRGB); }
        const paletteType = this.palette?.getProperty(PROPERTY_IDS.PALETTE_TYPE);
        if (this.paintingAction === 'palette') {
            if (paletteType === 0) { selectPaletteColor(this.player, this.palette, this.point, 0, this.cardinalDirection); }
            else if (paletteType === 1) { selectPaletteColor(this.player, this.palette, { x: getViewPointHorizontal(this.player, this.cardinalDirection, 60, 256), y: this.point.y }, 1); }
        }
    }
    
    manageShapes() {
        if (this.shapeStep > 0) {
            if (!this.shapeBox?.isValid()) { this.shapeBox = spawnShapeBox(this.easel, this.cardinalDirection); this.shapeBox?.triggerEvent(this.selectedShape === allShapes[0] ? EVENTS.LINE : EVENTS.SQUARE); }
            if (this.shapeBox?.isValid()) {
                this.shapeBox.setProperty(PROPERTY_IDS.FROM_X, this.shapeStep === 1 ? this.point.x : this.shapePoint.from.x);
                this.shapeBox.setProperty(PROPERTY_IDS.FROM_Y, this.shapeStep === 1 ? this.point.y : this.shapePoint.from.y);
                this.shapeBox.setProperty(PROPERTY_IDS.TO_X, this.point.x);
                this.shapeBox.setProperty(PROPERTY_IDS.TO_Y, this.point.y);
            }
        }
        if (this.shapeStep === 0 && this.paintingAction !== 'shape_step_0') {
            this.resetActions('shape_step_0');
            ShapesForm().show(this.player).then(({ canceled, selection }) => {
                if (canceled || selection === allShapes.length) { this.shapeStep = -1; return; }
                this.selectedShape = allShapes[selection]; this.shapeStep = 1;
            });
        }
        else if (this.shapeStep === 1) {
            if (this.paintingAction === 'shape_step_1') { 
                this.shapePoint.from = this.point; this.shapeStep = 2;
            }
        }
        else if (this.shapeStep === 2) {
            if (this.paintingAction === 'shape_step_2') {
                this.shapeStep = 1;
                this.shapePoint.to = this.point;
                const shape = this.selectedShape.split('|');
                createShape(this.player, this.pixels, this.colorRGB, shape[0], this.shapePoint.from, this.shapePoint.to, shape[1] === 'true');
                system.run(() => { this.updateSteps('shape_step_0'); this.shapeStep = 1; });
            }
        }
    }
    
    manageExposureTool() {
        if (this.exposureStep === 0 && this.paintingAction !== 'exposure_tool') {
            this.exposureStep = 1;
            ExposureForm(this.exposureValue < 0 ? this.exposureValue*-1 : this.exposureValue, this.lightenOrDarken).show(this.player).then(({ canceled, formValues }) => {
                if (canceled) { this.exposureStep = -1; return; }
                this.lightenOrDarken = formValues[1];
                this.exposureValue = formValues[0] * (this.lightenOrDarken ? 1 : -1);
            });
        } else if (this.exposureStep >= 1) {
            if (this.paintingAction === 'exposure_tool') {
                if (!this.exposurePoints.some(point => point.x === this.point.x && point.y === this.point.y)) { paintPixel(this.player, this.point, this.pixels, changePixelBrightness(getPixelColor(this.player, this.point, this.pixels, this.colorRGB, false), this.exposureValue), 1, false, false); }
                this.exposureStep = Math.max(this.exposureStep, 2);
                this.exposurePoints.push(this.point);
            } else if (this.exposureStep === 2 && this.paintingAction === 'start_paint_mode') {
                this.updateSteps(); this.exposureStep = 1;
            }
        }
    }
    
    _initializePaintMode() {
        this.canvas.setProperty(PROPERTY_IDS.PAINT_MODE, true);
        this.player.onScreenDisplay.setHudVisibility(HudVisibility.Hide, [HudElement.HorseHealth]);
        if (PAINT_MODE_CONFIG.HIDE_ITEM_TEXT) this.player.onScreenDisplay.setHudVisibility(HudVisibility.Hide, [HudElement.ItemText]);
        if (PAINT_MODE_CONFIG.HIDE_CROSSHAIR) this.player.onScreenDisplay.setHudVisibility(HudVisibility.Hide, [HudElement.Crosshair]);
        this.updateSteps();
    }
    
    _stopPaintMode(runInterval) {
        if (this.canvas.isValid()) this.canvas.setProperty(PROPERTY_IDS.PAINT_MODE, false);
        if (this.player.isValid()) PaintModeManager.resetPlayer(this.player);
        if (this.playerControl.isValid()) this.playerControl.remove();
        if (runInterval) system.clearRun(runInterval);
    }
    
    managePaintMode(isSaved = false) {
        this._initializePaintMode();
        const runPaintMode = system.runInterval(() => {
            if (!this.canvas.isValid() || !this.player.isValid() || !this.easel.isValid() || !this.playerControl.isValid() || this.paintingAction === 'stop_paint_mode' || this.player.getGameMode() === GameMode.spectator) { this._stopPaintMode(runPaintMode); return; }
            this.point = getViewPoint(this.player, this.cardinalDirection, VIEW_POINTS.FOV.x, VIEW_POINTS.FOV.y, VIEW_POINTS.OFFSET.y);
            // this.player.onScreenDisplay.setActionBar('x: ' + this.point.x + ' / y: ' + this.point.y);
            const mainhand = PlayerManager.getItem(this.player, 'mainhand');
            this.brushSize = this.getBrushSize(mainhand);
            this.paintingAction = getPaintingAction(this.player);
            this.canvas.setProperty(PROPERTY_IDS.POSITION_X, this.point.x);
            this.canvas.setProperty(PROPERTY_IDS.POSITION_Y, this.point.y);
            this.canvas.setProperty(PROPERTY_IDS.BRUSH_SIZE, this.brushSize);
            if (isSaved) {
                if (mainhand?.typeId === ITEMS.EYEDROPPER) {
                    if (this.paintingAction === 'eyedropper') { OBJECTIVES.COLOR_RGB.setScore(this.player, getPixelColor(this.player, this.point, this.pixels, this.colorRGB)); }
                } else { this._stopPaintMode(runPaintMode); return; }
                this.colorRGB = OBJECTIVES.COLOR_RGB.getScore(this.player);
                this.canvas.setProperty(PROPERTY_IDS.COLOR_RGB, this.colorRGB); 
                return;
            }
            if (mainhand?.typeId === ITEMS.PALETTE) { this.managePalette(); }
            else if (this.palette?.isValid()) { choosePaletteColor(this.player, this.palette); this.resetActions(); return; }
            if (mainhand?.typeId === ITEMS.SHAPES) { this.manageShapes(); }
            else if (this.shapeStep !== 0) { this.resetActions(); return; }
            if (mainhand?.typeId === ITEMS.EXPOSURE_TOOL) { this.manageExposureTool() }
            else if (this.exposureStep !== 0) { this.exposureStep = 0 }
            switch (this.paintingAction) {
                case 'start_paint_mode': break;
                case 'brush': paintPixel(this.player, this.point, this.pixels, this.colorRGB, this.brushSize, PAINT_MODE_CONFIG.BRUSH_SOUND, PAINT_MODE_CONFIG.ALLOW_PARTICLES); break;
                case 'large_brush': paintPixel(this.player, this.point, this.pixels, this.colorRGB, this.brushSize, PAINT_MODE_CONFIG.BRUSH_SOUND, PAINT_MODE_CONFIG.ALLOW_PARTICLES); break;
                case 'paint_remover': clearPixel(this.point, this.pixels); break;
                case 'paint_bucket': fillPixels(this.player, this.point, this.pixels, this.colorRGB, PAINT_MODE_CONFIG.BUCKET_SOUND, PAINT_MODE_CONFIG.ALLOW_PARTICLES); break;
                case 'eyedropper': OBJECTIVES.COLOR_RGB.setScore(this.player, getPixelColor(this.player, this.point, this.pixels, this.colorRGB)); break;
                case 'update_steps': this.updateSteps(); break;
                case 'undo': this.undoRedo('undo'); break;
                case 'redo': this.undoRedo('redo'); break;
            }
            this.colorRGB = OBJECTIVES.COLOR_RGB.getScore(this.player);
            this.canvas.setProperty(PROPERTY_IDS.COLOR_RGB, this.colorRGB);
        });
    }
}

export default PaintModeManager;